/****** WIN32 ONLY ******/
/* Volumetric fog using EXT_fog_coord extension */
/* An OpenGL demo by A.P.Gardner */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include "GL_EXT_fog_coord.h"

/* definitions */

#define NUM_PANELS 16	/* number of panels in scene */

typedef struct
{
	float x;
	float y;
	float z;
} world_vector;		/* vertex data */

/* variables */

GLuint texName[2];		/* store of texture references */
int ext_fog_coord = 0;	/* flag indicating fog coord support */
PFNGLFOGCOORDFEXTPROC glFogCoordfEXT = NULL;
	/* vertex data for the panels */
world_vector vertices[NUM_PANELS*4] = {
	{-15.0, 10.0, 0.0},{-5.0, 10.0, 0.0},{-5.0, 20.0, 0.0},{-15.0, 20.0, 0.0}, /* back wall */
	{-5.0, 10.0, 0.0},{5.0, 10.0, 0.0},{5.0, 20.0, 0.0},{-5.0, 20.0, 0.0},
	{5.0, 10.0, 0.0},{15.0, 10.0, 0.0},{15.0, 20.0, 0.0},{5.0, 20.0, 0.0},
	{-15.0, 0.0, 0.0},{-5.0, 0.0, 0.0},{-5.0, 10.0, 0.0},{-15.0, 10.0, 0.0},
	{-5.0, 0.0, 0.0},{5.0, 0.0, 0.0},{5.0, 10.0, 0.0},{-5.0, 10.0, 0.0},
	{5.0, 0.0, 0.0},{15.0, 0.0, 0.0},{15.0, 10.0, 0.0},{5.0, 10.0, 0.0},
	{-5.0, -10.0, 0.0},{5.0, -10.0, 0.0},{5.0, 0.0, 0.0},{-5.0, 0.0, 0.0}, /* foggy bit */
	{-5.0, -10.0, 10.0},{-5.0, -10.0, 0.0},{-5.0, 0.0, 0.0},{-5.0, 0.0, 10.0},
	{-5.0, -10.0, 20.0},{-5.0, -10.0, 10.0},{-5.0, 0.0, 10.0},{-5.0, 0.0, 20.0},
	{5.0, -10.0, 0.0},{5.0, -10.0, 10.0},{5.0, 0.0, 10.0},{5.0, 0.0, 0.0},
	{5.0, -10.0, 10.0},{5.0, -10.0, 20.0},{5.0, 0.0, 20.0},{5.0, 0.0, 10.0},
	{-5.0, -10.0, 10.0},{5.0, -10.0, 10.0},{5.0, -10.0, 0.0},{-5.0, -10.0, 0.0},
	{-15.0, 0.0, 10.0},{-5.0, 0.0, 10.0},{-5.0, 0.0, 0.0},{-15.0, 0.0, 0.0}, /* floor */
	{5.0, 0.0, 10.0},{15.0, 0.0, 10.0},{15.0, 0.0, 0.0},{5.0, 0.0, 0.0},
	{-15.0, 0.0, 20.0},{-5.0, 0.0, 20.0},{-5.0, 0.0, 10.0},{-15.0, 0.0, 10.0},
	{5.0, 0.0, 20.0},{15.0, 0.0, 20.0},{15.0, 0.0, 10.0},{5.0, 0.0, 10.0}
};
	/* texture references for the panels */
int panel_texture[NUM_PANELS] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0};
	/* area references for the panels */
int fog_area[NUM_PANELS] = {0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0};

/**************************************************************************/

/* Function to detect and set up fog coord support */

int fog_coord_init()
{
	const GLubyte *glextstring;
	char extName[] = "EXT_fog_coord";

	/* read extensions string */

	glextstring = glGetString(GL_EXTENSIONS);
	if (glextstring == NULL)
	{
		printf("Fatal error, no extensions string!\n");
		return 0;
	}

	/* check for EXT_fog_coord */

	if (strstr(glextstring, extName) == NULL)
	{
		printf("EXT_fog_coord not found, you may need a newer driver!\n");
		return 0;
	}

	/* set up proc addresses & multitexture flag */

    glFogCoordfEXT = (void*)wglGetProcAddress("glFogCoordfEXT");
	ext_fog_coord = 1;

	return 1;
}

/* Function to load a texture from a bmp file */

void makeTextureImage(char filename[], GLuint tex_name)
{
	long data_start = 54;			/* start of data section in files wall.bmp & light.bmp */
									/* NB header size of bmp files is not fixed, refer to M$ documentation */
	int size = 256;					/* width/height of textures */
	unsigned char ch, cho[4];		/* temp file input buffers */
	FILE *fileptr;					/* pointer to bmp file */
	GLubyte image256[256][256][3];	/* storage for 256x256 texture */
	int n, m, i, j;					/* counters */

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the texture data */

	for (i=0; i<size; i++)
	{
		for (j=0; j<size; j++)
		{
			m=0;
			while(!feof(fileptr) && m < 3)
			{
				cho[m] = fgetc(fileptr);
				m++;
			}
			image256[i][j][0] = cho[2];	/* bmp using BGR values, hence the swap */
			image256[i][j][1] = cho[1];
			image256[i][j][2] = cho[0];
		}
	}

	/* close the file */

	fclose(fileptr);

	/* create texture object */
	
	glBindTexture(GL_TEXTURE_2D, texName[tex_name]);

	/* set up wrap & filter params */

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	/* load the image */

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, size, size,
				 0, GL_RGB, GL_UNSIGNED_BYTE, image256);
}

/* Function to initialise params */

void init(void)
{
	GLfloat fogColor[4] = {0.6, 0.2, 0.0, 1.0};	/* volumetric fog colour */

	/* set background colour */

	glClearColor(0.0, 0.0, 0.0, 0.0);

	/* initialise viewing parameters */

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
    gluPerspective(50, 1.0, 0.2, 40);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* generate texture names */

	texName[0] = 0;
	texName[1] = 0;
	glGenTextures(2, texName);

	/* load textures */

	makeTextureImage("floor.bmp", 0);
	makeTextureImage("wall.bmp", 1);

	/* enable texturing & set texturing function */

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	/* set up fog params */

	glEnable(GL_FOG);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogfv(GL_FOG_COLOR, fogColor);
	glFogf(GL_FOG_START, 0.0);
	glFogf(GL_FOG_END, 10.0);
	if (ext_fog_coord)
		glFogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT);
}

/* Function to clean up */

void cleanup(void)
{
	/* delete texture objects */

	glDeleteTextures(2, texName);
}

/* Function to set fog colour for the current vertex */

void set_fog_colour(int p, int v)
{
	float z;	/* distance of vertex from edge of fog volume */

	/* calculate depth through fog */

	if (fog_area[p])
	{
		z = 0.0 - vertices[v].y;
	}
	else
	{
		z = 0.0;
	}

	/* set depth for this coord */

	glFogCoordfEXT(z);
}

/* Function to draw a panel */

void draw_panel(int ref)
{
	int v;		/* store of next vertex ref */

	/* check for no extension support */

	if (!ext_fog_coord)
		return ;

	/* bind texture */

	glBindTexture(GL_TEXTURE_2D, texName[panel_texture[ref]]);

	/* draw panel */

	v = ref * 4;
	glBegin(GL_QUADS);
		set_fog_colour(ref, v);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z); v++;
		set_fog_colour(ref, v);
		glTexCoord2f(1.0, 0.0);
		glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z); v++;
		set_fog_colour(ref, v);
		glTexCoord2f(1.0, 1.0);
		glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z); v++;
		set_fog_colour(ref, v);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z);
	glEnd();
}

/* Function to draw display */

void display(void)
{
	int n;		/* counter */

	/* clear all pixels in colour buffer */

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* position viewer */

	glLoadIdentity();
	glTranslatef(0.0, -3.0, -32.0);
	glRotatef(10.0, 1.0, 0.0, 0.0);

	/* draw the panels */

	for (n=0; n<NUM_PANELS; n++)
		draw_panel(n);

	/* process all buffered OpenGL commands */

	glFlush();
}

/* Main function */

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_SINGLE | GLUT_DEPTH |GLUT_RGB);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("OpenGL Demo - Volumetric Fog (EXT_fog_coord)");
	fog_coord_init();
	init();
	glutDisplayFunc(display);
	glutMainLoop();
	cleanup();

	return 0;
}